@@ -3,9 +3,9 @@
  *
  * Common functions for SAKURA and TOMOYO.
  *
- * Copyright (C) 2005  NTT DATA CORPORATION
+ * Copyright (C) 2005-2006  NTT DATA CORPORATION
  *
- * Version: 1.0.1 2005/11/25
+ * Version: 1.1.2   2006/06/02
  *
  * This file is applicable to both 2.4.30 and 2.6.11 and later.
  * See README.ccs for ChangeLog.
@@ -31,58 +31,52 @@ static const int lookup_flags = LOOKUP_FOLLOW | LOOKUP_POSITIVE;
 #include <linux/realpath.h>
 #include <linux/ccs_common.h>
 #include <linux/ccs_proc.h>
-
-/*************************  PROTOTYPES  *************************/
-
-static int AddPolicyManagerPolicy(char *exe, void **dummy);
-static int IsPolicyManager(void);
-static int LoadPolicyManager(void);
-#ifdef CONFIG_TOMOYO
-static int AddDomainPolicy(char *data, void **domain);
-static int LoadDomainPolicy(void);
-static int ReadDomainPolicy(IO_BUFFER *head);
-static int AddExceptionPolicy(char *data, void **dummy);
-static int LoadExceptionPolicy(void);
-static int ReadExceptionPolicy(IO_BUFFER *head);
-#endif
-#ifdef CONFIG_SAKURA
-static int LoadSystemPolicy(void);
-static int ReadSystemPolicy(IO_BUFFER *head);
-#endif
-static int ReadMemoryCounter(IO_BUFFER *head);
-static int CopyToUser(IO_BUFFER *head, char __user * buffer, int buffer_len);
+#include <linux/tomoyo.h>
 
 /*************************  VARIABLES  *************************/
 
 /* /sbin/init started? */
-extern int sbin_init_started;
+int sbin_init_started = 0;
+
+const char *ccs_log_level = KERN_DEBUG;
 
 static struct {
-	const unsigned int assertion_index;
 	const char *keyword;
 	unsigned int current_value;
 	const unsigned int max_value;
 } ccs_control_array[] = {
-	{ CCS_TOMOYO_MAC_FOR_FILE,        "MAC_FOR_FILE",        0, 3 },              /* domain_policy.txt */
-	{ CCS_TOMOYO_MAC_FOR_BINDPORT,    "MAC_FOR_BINDPORT",    0, 3 },              /* domain_policy.txt */
-	{ CCS_TOMOYO_MAC_FOR_CONNECTPORT, "MAC_FOR_CONNECTPORT", 0, 3 },              /* domain_policy.txt */
-	{ CCS_TOMOYO_MAC_FOR_SIGNAL,      "MAC_FOR_SIGNAL",      0, 3 },              /* domain_policy.txt */
-	{ CCS_SAKURA_DENY_CONCEAL_MOUNT,  "DENY_CONCEAL_MOUNT",  0, 3 },
-	{ CCS_SAKURA_RESTRICT_CHROOT,     "RESTRICT_CHROOT",     0, 3 },              /* system_policy.txt */
-	{ CCS_SAKURA_RESTRICT_MOUNT,      "RESTRICT_MOUNT",      0, 3 },              /* system_policy.txt */
-	{ CCS_SAKURA_RESTRICT_UNMOUNT,    "RESTRICT_UNMOUNT",    0, 3 },              /* system_policy.txt */
-	{ CCS_SAKURA_DENY_PIVOT_ROOT,     "DENY_PIVOT_ROOT",     0, 3 },
-	{ CCS_SAKURA_TRACE_READONLY,      "TRACE_READONLY",      0, 1 },
-	{ CCS_SAKURA_RESTRICT_AUTOBIND,   "RESTRICT_AUTOBIND",   0, 1 },              /* system_policy.txt */
-	{ CCS_TOMOYO_MAX_ACCEPT_FILES,    "MAX_ACCEPT_FILES",    MAX_ACCEPT_FILES, UINT_MAX },
-	{ CCS_TOMOYO_MAX_GRANT_LOG,       "MAX_GRANT_LOG",       MAX_GRANT_LOG, MAX_GRANT_LOG },
-	{ CCS_TOMOYO_MAX_REJECT_LOG,      "MAX_REJECT_LOG",      MAX_REJECT_LOG, MAX_REJECT_LOG },
-	{ CCS_TOMOYO_VERBOSE,             "TOMOYO_VERBOSE",      1, 1 },
-	{ 0, NULL, 0, 0 }
+	{ "MAC_FOR_FILE",        0, 3 },                      // CCS_TOMOYO_MAC_FOR_FILE
+	{ "MAC_FOR_BINDPORT",    0, 3 },                      // CCS_TOMOYO_MAC_FOR_BINDPORT
+	{ "MAC_FOR_CONNECTPORT", 0, 3 },                      // CCS_TOMOYO_MAC_FOR_CONNECTPORT
+	{ "MAC_FOR_SIGNAL",      0, 3 },                      // CCS_TOMOYO_MAC_FOR_SIGNAL
+	{ "DENY_CONCEAL_MOUNT",  0, 3 },                      // CCS_SAKURA_DENY_CONCEAL_MOUNT
+	{ "RESTRICT_CHROOT",     0, 3 },                      // CCS_SAKURA_RESTRICT_CHROOT
+	{ "RESTRICT_MOUNT",      0, 3 },                      // CCS_SAKURA_RESTRICT_MOUNT
+	{ "RESTRICT_UNMOUNT",    0, 3 },                      // CCS_SAKURA_RESTRICT_UNMOUNT
+	{ "DENY_PIVOT_ROOT",     0, 3 },                      // CCS_SAKURA_DENY_PIVOT_ROOT
+	{ "TRACE_READONLY",      0, 1 },                      // CCS_SAKURA_TRACE_READONLY
+	{ "RESTRICT_AUTOBIND",   0, 1 },                      // CCS_SAKURA_RESTRICT_AUTOBIND
+	{ "MAX_ACCEPT_FILES",    MAX_ACCEPT_FILES, INT_MAX }, // CCS_TOMOYO_MAX_ACCEPT_FILES
+	{ "MAX_GRANT_LOG",       MAX_GRANT_LOG, INT_MAX },    // CCS_TOMOYO_MAX_GRANT_LOG
+	{ "MAX_REJECT_LOG",      MAX_REJECT_LOG, INT_MAX },   // CCS_TOMOYO_MAX_REJECT_LOG
+	{ "TOMOYO_VERBOSE",      1, 1 },                      // CCS_TOMOYO_VERBOSE
+	{ "MAX_ENFORCE_GRACE",   0, INT_MAX },                // CCS_MAX_ENFORCE_GRACE
+	{ NULL, 0, 0 }
 };
 
 /*************************  UTILITY FUNCTIONS  *************************/
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
+/* Taken from kernel 2.6 */
+void *kzalloc(const size_t size, const int flags)
+{
+	void *ret = kmalloc(size, flags);
+	if (ret) memset(ret, 0, size);
+	return ret;
+}
+#endif
+
+#ifdef CONFIG_TOMOYO
 static int __init TOMOYO_Quiet_Setup(char *str)
 {
 	ccs_control_array[CCS_TOMOYO_VERBOSE].current_value = 0;
@@ -90,9 +84,10 @@ static int __init TOMOYO_Quiet_Setup(char *str)
 }
 
 __setup("TOMOYO_QUIET", TOMOYO_Quiet_Setup);
+#endif
 
 /* Am I root? */
-int isRoot(void)
+static int isRoot(void)
 {
 	return !current->uid && !current->euid;
 }
@@ -184,8 +179,9 @@ int IsCorrectPath(const char *filename, const int may_contain_pattern)
  */
 int PathMatchesToPattern(const char *pathname, const char *pattern)
 {
+	const char *org_pathname = pathname;
 	if (!pathname || !pattern) return 0;
-	/* if pattern doesn't contain \ , I can use strcmp(). */
+	/* if pattern doesn't contain '\', I can use strcmp(). */
 	if (!strchr(pattern, '\\')) return !strcmp(pathname, pattern);
 	while (*pathname && *pattern) {
 		if (*pattern == '\\') {
@@ -237,8 +233,8 @@ int PathMatchesToPattern(const char *pathname, const char *pattern)
 			pathname++; pattern++;
 		}
 	}
-	/* Skip trailing '\*'s in pattern string. */
-	if (!*pathname) while (*pattern == '\\' && *(pattern + 1) == '*') pattern += 2;
+	/* Skip trailing '\*' in pattern string, if pathname is not a directory. */
+	if (!*pathname && *pattern == '\\' && *(pattern + 1) == '*' && org_pathname < pathname && *(pathname - 1) != '/') pattern += 2;
 	return (!*pathname && !*pattern);
 }
 
@@ -252,8 +248,7 @@ int ReadFileAndProcess(struct file *file, int (*func) (char *, void **))
 	int len;
 	unsigned long offset = 0;
 	if (!file || !func) return -EINVAL;
-	if ((buffer = kmalloc(PAGE_SIZE * 2, GFP_KERNEL)) == NULL) return -ENOMEM;
-	memset(buffer, 0, PAGE_SIZE * 2);
+	if ((buffer = kzalloc(PAGE_SIZE * 2, GFP_KERNEL)) == NULL) return -ENOMEM;
 	/* Reads one line and process it. Ends when no more '\n' found. */
 	while ((len = kernel_read(file, offset, buffer, PAGE_SIZE * 2)) > 0 && (cp = memchr(buffer, '\n', len)) != NULL) {
 		*cp = '\0';
@@ -293,9 +288,8 @@ const char *GetEXE(void)
 		struct vm_area_struct *vma = current->mm->mmap;
 		while (vma) {
 			if ((vma->vm_flags & VM_EXECUTABLE) && vma->vm_file) {
-				char *buf = kmalloc(PAGE_SIZE, GFP_KERNEL);
+				char *buf = kzalloc(PAGE_SIZE, GFP_KERNEL);
 				if (buf == NULL) return NULL;
-				memset(buf, 0, PAGE_SIZE);
 				if (realpath_from_dentry(vma->vm_file->f_dentry, vma->vm_file->f_vfsmnt, buf, PAGE_SIZE - 1) == 0) return (const char *) buf;
 				kfree(buf); return NULL;
 			}
@@ -320,16 +314,21 @@ unsigned int TomoyoVerboseMode(void)
 /* Default mode is "disable all mandatory access control functions". */
 /* You must create profiles and choose one using CCS= options to the kernel command line. */
 static unsigned int ccs_profile_index = 0;
-static int skip_domain_policy_loading = 0;
+static int panic_if_load_profile_fails = 0;
 
 static int __init CCS_Setup(char *str)
 {
+	panic_if_load_profile_fails = 1;
 	if (sscanf(str, "%u", &ccs_profile_index) != 1) printk("Invalid value for CCS= .");
 	return 0;
 }
 
 __setup("CCS=", CCS_Setup);
 
+#ifdef CONFIG_TOMOYO
+
+static int skip_domain_policy_loading = 0;
+
 static int __init TOMOYO_NOLOAD_Setup(char *str)
 {
 	skip_domain_policy_loading = 1;
@@ -338,6 +337,7 @@ static int __init TOMOYO_NOLOAD_Setup(char *str)
 }
 
 __setup("TOMOYO_NOLOAD", TOMOYO_NOLOAD_Setup);
+#endif
 
 /* Check whether the given access control is enabled. */
 unsigned int CheckCCSFlags(const unsigned int index)
@@ -351,30 +351,27 @@ unsigned int CheckCCSFlags(const unsigned int index)
 /* Check whether the given access control is enforce mode. */
 unsigned int CheckCCSEnforce(const unsigned int index)
 {
-	if (index < (sizeof(ccs_control_array) / sizeof(ccs_control_array[0])) - 1)
-		return ccs_control_array[index].current_value == 3;
-	printk("%s: Index %u is out of range. Fix the kernel source.\n", __FUNCTION__, index);
-	return 0;
+	return CheckCCSFlags(index) == 3;
 }
 
 /* Check whether the given access control is accept mode. */
 unsigned int CheckCCSAccept(const unsigned int index)
 {
-	if (index < (sizeof(ccs_control_array) / sizeof(ccs_control_array[0])) - 1)
-		return ccs_control_array[index].current_value == 1;
-	printk("%s: Index %u is out of range. Fix the kernel source.\n", __FUNCTION__, index);
-	return 0;
+	return CheckCCSFlags(index) == 1;
 }
 
-unsigned int GetMaxAutoAppendFiles(void) {
+unsigned int GetMaxAutoAppendFiles(void)
+{
 	return ccs_control_array[CCS_TOMOYO_MAX_ACCEPT_FILES].current_value;
 }
 
-unsigned int GetMaxGrantLog(void) {
+unsigned int GetMaxGrantLog(void)
+{
 	return ccs_control_array[CCS_TOMOYO_MAX_GRANT_LOG].current_value;
 }
 
-unsigned int GetMaxRejectLog(void) {
+unsigned int GetMaxRejectLog(void)
+{
 	return ccs_control_array[CCS_TOMOYO_MAX_REJECT_LOG].current_value;
 }
 
@@ -393,11 +390,10 @@ static int SetStatus(char *data, void **dummy)
 	}
 #endif
 	for (i = 0; ccs_control_array[i].keyword; i++) {
-		if (strcmp(data, ccs_control_array[i].keyword) == 0) {
-			if (value > ccs_control_array[i].max_value) value = ccs_control_array[i].max_value;
-			ccs_control_array[i].current_value = value;
-			break;
-		}
+		if (strcmp(data, ccs_control_array[i].keyword)) continue;
+		if (value > ccs_control_array[i].max_value) value = ccs_control_array[i].max_value;
+		ccs_control_array[i].current_value = value;
+		break;
 	}
 	return ccs_control_array[i].keyword ? 0 : -EINVAL;
 }
@@ -405,7 +401,7 @@ static int SetStatus(char *data, void **dummy)
 static int LoadProfile(void)
 {
 	struct file *f;
-	static char buffer[64];
+	char buffer[64];
 	memset(buffer, 0, sizeof(buffer));
 	snprintf(buffer, sizeof(buffer) - 1, "/root/security/profile%u.txt", ccs_profile_index);
 	f = filp_open(buffer, O_RDONLY, 0600);
@@ -415,6 +411,7 @@ static int LoadProfile(void)
 		printk("%s: Loading profiles done.\n", __FUNCTION__);
 	} else {
 		printk("%s: Can't load %s\n", __FUNCTION__, buffer);
+		if (panic_if_load_profile_fails) panic("%s: Failed to load profile.\n", __FUNCTION__);
 	}
 	return 0;
 }
@@ -426,8 +423,51 @@ static int ReadStatus(IO_BUFFER *head)
 		if (!isRoot()) return -EPERM;
 		if (!head->read_var2) {
 			for (i = head->read_step; ccs_control_array[i].keyword; i++) {
-				if (io_printf(head, "%s=%u\n", ccs_control_array[i].keyword, ccs_control_array[i].current_value)) break;
 				head->read_step = i;
+				switch (i) {
+				case -1: // Dummy
+#ifndef CONFIG_TOMOYO_MAC_FOR_FILE
+				case CCS_TOMOYO_MAC_FOR_FILE:
+				case CCS_TOMOYO_MAX_ACCEPT_FILES:
+#endif
+#ifndef CONFIG_TOMOYO_MAC_FOR_BINDPORT
+				case CCS_TOMOYO_MAC_FOR_BINDPORT:
+#endif
+#ifndef CONFIG_TOMOYO_MAC_FOR_CONNECTPORT
+				case CCS_TOMOYO_MAC_FOR_CONNECTPORT:
+#endif
+#ifndef CONFIG_TOMOYO_MAC_FOR_SIGNAL
+				case CCS_TOMOYO_MAC_FOR_SIGNAL:
+#endif
+#ifndef CONFIG_SAKURA_DENY_CONCEAL_MOUNT
+				case CCS_SAKURA_DENY_CONCEAL_MOUNT:
+#endif
+#ifndef CONFIG_SAKURA_RESTRICT_CHROOT
+				case CCS_SAKURA_RESTRICT_CHROOT:
+#endif
+#ifndef CONFIG_SAKURA_RESTRICT_MOUNT
+				case CCS_SAKURA_RESTRICT_MOUNT:
+#endif
+#ifndef CONFIG_SAKURA_RESTRICT_UNMOUNT
+				case CCS_SAKURA_RESTRICT_UNMOUNT:
+#endif
+#ifndef CONFIG_SAKURA_DENY_PIVOT_ROOT
+				case CCS_SAKURA_DENY_PIVOT_ROOT:
+#endif
+#ifndef CONFIG_SAKURA_TRACE_READONLY
+				case CCS_SAKURA_TRACE_READONLY:
+#endif
+#ifndef CONFIG_SAKURA_RESTRICT_AUTOBIND
+				case CCS_SAKURA_RESTRICT_AUTOBIND:
+#endif
+#ifndef CONFIG_TOMOYO
+				case CCS_TOMOYO_MAX_GRANT_LOG:
+				case CCS_TOMOYO_MAX_REJECT_LOG:
+				case CCS_TOMOYO_VERBOSE:
+#endif
+					continue;
+				}
+				if (io_printf(head, "%s=%u\n", ccs_control_array[i].keyword, ccs_control_array[i].current_value)) break;
 			}
 			if (!ccs_control_array[i].keyword) {
 				head->read_var2 = (void *) 1;
@@ -470,8 +510,6 @@ static int AddPolicyManagerPolicy(char *exe, void **dummy)
 	/* I don't want to add if it was already added. */
 	for (ptr = policy_manager_list.next; ptr; ptr = ptr->next) if (strcmp(ptr->exe, exe) == 0) return 0;
 	if ((saved_exe = SaveName(exe)) == NULL || (new_entry = (POLICY_MANAGER_ENTRY *) alloc_element(sizeof(POLICY_MANAGER_ENTRY))) == NULL) return -ENOMEM;
-	memset(new_entry, 0, sizeof(POLICY_MANAGER_ENTRY));
-	new_entry->next = NULL;
 	new_entry->exe = saved_exe;
 	/***** CRITICAL SECTION START *****/
 	spin_lock(&lock);
@@ -486,11 +524,18 @@ static int IsPolicyManager(void)
 {
 	POLICY_MANAGER_ENTRY *ptr;
 	const char *exe;
-	if (!policy_manager_list.next) return 0;
 	if ((exe = GetEXE()) == NULL) return 0;
 	for (ptr = policy_manager_list.next; ptr; ptr = ptr->next) {
 		if (strcmp(exe, ptr->exe) == 0) break;
 	}
+	if (!ptr) { /* Reduce error messages. */
+		static pid_t last_pid = 0;
+		const pid_t pid = current->pid;
+		if (last_pid != pid) {
+			printk("%s is not permitted to update policies.\n", exe);
+			last_pid = pid;
+		}
+	}
 	kfree(exe);
 	return ptr ? 1 : 0;
 }
@@ -507,38 +552,72 @@ static int LoadPolicyManager(void)
 	return 0;
 }
 
-/*************************  DOMAIN POLICY HANDLER  *************************/
-
 #ifdef CONFIG_TOMOYO
 
+#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+
+/*************************  PERMISSION MAPPING HANDLER  *************************/
+
+/* Definition file for permission mapping. */
+#define POLICY_MAPPING_FILE           "/root/security/mapping.txt"
+
+static int LoadPermissionMapping(void)
+{
+	struct file *f = filp_open(POLICY_MAPPING_FILE, O_RDONLY, 0600);
+	if (!IS_ERR(f)) {
+		if (f->f_dentry->d_inode->i_size > 0 && ReadFileAndProcess(f, SetPermissionMapping) < 0) panic("%s: Failed to load.\n", __FUNCTION__);
+		filp_close(f, NULL);
+		printk("%s: Loading permission mappings done.\n", __FUNCTION__);
+	}
+	return 0;
+}
+#endif
+
+/*************************  DOMAIN POLICY HANDLER  *************************/
+
 /* Definition file for domains policy. */
 #define DOMAIN_POLICY_FILE           "/root/security/domain_policy.txt"
 
 static int AddDomainPolicy(char *data, void **domain)
 {
+	int is_delete = 0, is_select = 0;
 	if (!isRoot()) return -EPERM;
+	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
+		data += KEYWORD_DELETE_LEN;
+		is_delete = 1;
+	} else if (strncmp(data, KEYWORD_SELECT, KEYWORD_SELECT_LEN) == 0) {
+		data += KEYWORD_SELECT_LEN;
+		is_select = 1;
+	}
 	if (IsDomainDef(data)) {
-		*domain = FindOrAssignNewDomain(data);
+		if (is_delete) {
+			DeleteDomain(data);
+			*domain = NULL;
+		} else if (is_select) {
+			*domain = FindDomain(data);
+		} else {
+			*domain = FindOrAssignNewDomain(data);
+		}
 		return 0;
 #ifdef CONFIG_TOMOYO_MAC_FOR_CAPABILITY
 	} else if (strncmp(data, KEYWORD_ALLOW_CAPABILITY, KEYWORD_ALLOW_CAPABILITY_LEN) == 0) {
-		return AddCapabilityPolicy(data + KEYWORD_ALLOW_CAPABILITY_LEN, domain);
+		return AddCapabilityPolicy(data + KEYWORD_ALLOW_CAPABILITY_LEN, (struct domain_info *) *domain, is_delete);
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_BINDPORT
 	} else if (strncmp(data, KEYWORD_ALLOW_BIND, KEYWORD_ALLOW_BIND_LEN) == 0) {
-		return AddBindPolicy(data + KEYWORD_ALLOW_BIND_LEN, domain);
+		return AddBindPolicy(data + KEYWORD_ALLOW_BIND_LEN, (struct domain_info *) *domain, is_delete);
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_CONNECTPORT
 	} else if (strncmp(data, KEYWORD_ALLOW_CONNECT, KEYWORD_ALLOW_CONNECT_LEN) == 0) {
-		return AddConnectPolicy(data + KEYWORD_ALLOW_CONNECT_LEN, domain);
+		return AddConnectPolicy(data + KEYWORD_ALLOW_CONNECT_LEN, (struct domain_info *) *domain, is_delete);
 #endif
 #ifdef CONFIG_TOMOYO_MAC_FOR_SIGNAL
 	} else if (strncmp(data, KEYWORD_ALLOW_SIGNAL, KEYWORD_ALLOW_SIGNAL_LEN) == 0) {
-		return AddSignalPolicy(data + KEYWORD_ALLOW_SIGNAL_LEN, domain);
+		return AddSignalPolicy(data + KEYWORD_ALLOW_SIGNAL_LEN, (struct domain_info *) *domain, is_delete);
 #endif
 	} else {
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
-		return AddFilePolicy(data, domain);
+		return AddFilePolicy(data, (struct domain_info *) *domain, is_delete);
 #endif
 	}
 	return -EINVAL;
@@ -567,46 +646,72 @@ static int ReadDomainPolicy(IO_BUFFER *head)
 		case 1: goto step1;
 		case 2: goto step2;
 		case 3: goto step3;
-		case 4: goto step4;
-		case 5: goto step5;
-		case 6: goto step6;
-		case 7: goto step7;
 		default: return -EINVAL;
 		}
 		if (!isRoot()) return -EPERM;
 		for (domain = &KERNEL_DOMAIN; domain; domain = domain->next) {
+			struct acl_info *ptr;
 			if (domain->attribute & DOMAIN_ATTRIBUTE_DELETED) continue;
 			head->read_var1 = (void *) domain;
 			head->read_var2 = NULL; head->read_step = 1;
 		step1:
 			if (io_printf(head, "%s\n\n", domain->domainname)) break;
-			head->read_var2 = NULL; head->read_step = 2;
+			head->read_var2 = (void *) domain->first_acl_ptr; head->read_step = 2;
 		step2:
+			for (ptr = (struct acl_info *) head->read_var2; ptr; ptr = ptr->next) {
+				const unsigned int acl_type = GET_ACL_TYPE(ptr->type_hash);
+				head->read_var2 = (void *) ptr;
+				if (0) {
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
-			if (ReadFilePolicy(head)) break;
+				} else if (acl_type == TYPE_FILE_ACL) {
+					if (io_printf(head, "%d %s \n", ((FILE_ACL_RECORD *) ptr)->perm, ((FILE_ACL_RECORD *) ptr)->filename)) break;
+				} else if (acl_type == TYPE_DIR_ACL) {
+					if (io_printf(head, "%d %s \n", 7, ((DIR_ACL_RECORD *) ptr)->filename)) break;
 #endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_CAPABILITY
+				} else if (acl_type == TYPE_CAPABILITY_ACL) {
+					if (io_printf(head, KEYWORD_ALLOW_CAPABILITY "%s\n", capability2keyword(GET_ACL_HASH(ptr->type_hash)))) break;
+#endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_BINDPORT
+				} else if (acl_type == TYPE_BIND_ACL) {
+					const int is_stream = GET_ACL_HASH(ptr->type_hash);
+					const unsigned short int min_port = ((NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((NETWORK_ACL_RECORD *) ptr)->max_port; 
+					if (min_port != max_port) {
+						if (io_printf(head, KEYWORD_ALLOW_BIND "%s/%u-%u\n", is_stream ? "TCP" : "UDP", min_port, max_port)) break;
+					} else {
+						if (io_printf(head, KEYWORD_ALLOW_BIND "%s/%u\n", is_stream ? "TCP" : "UDP", min_port)) break;
+					}
+#endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_CONNECTPORT
+				} else if (acl_type == TYPE_CONNECT_ACL) {
+					const int is_stream = GET_ACL_HASH(ptr->type_hash);
+					const unsigned short int min_port = ((NETWORK_ACL_RECORD *) ptr)->min_port, max_port = ((NETWORK_ACL_RECORD *) ptr)->max_port; 
+					if (min_port != max_port) {
+						if (io_printf(head, KEYWORD_ALLOW_CONNECT "%s/%u-%u\n", is_stream ? "TCP" : "UDP", min_port, max_port)) break;
+					} else {
+						if (io_printf(head, KEYWORD_ALLOW_CONNECT "%s/%u\n", is_stream ? "TCP" : "UDP", min_port)) break;
+					}
+#endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_SIGNAL
+				} else if (acl_type == TYPE_SIGNAL_ACL) {
+					if (io_printf(head, KEYWORD_ALLOW_SIGNAL "%u %s\n", GET_ACL_HASH(ptr->type_hash), ((SIGNAL_ACL_RECORD *) ptr)->domainname)) break;
+#endif
+#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+				} else {
+					const char *keyword = acltype2keyword(acl_type);
+					if (keyword) {
+						if (acltype2paths(acl_type) == 2) {
+							if (io_printf(head, "allow_%s %s %s \n", keyword, ((DOUBLE_ACL_RECORD *) ptr)->filename1, ((DOUBLE_ACL_RECORD *) ptr)->filename2)) break;
+						} else {
+							if (io_printf(head, "allow_%s %s \n", keyword, ((SINGLE_ACL_RECORD *) ptr)->filename)) break;
+						}
+					}
+#endif
+				}
+			}
+			if (ptr) break;
 			head->read_var2 = NULL; head->read_step = 3;
 		step3:
-#ifdef CONFIG_TOMOYO_MAC_FOR_CAPABILITY
-			if (ReadCapabilityPolicy(head)) break;
-#endif
-			head->read_var2 = NULL; head->read_step = 4;
-		step4:
-#ifdef CONFIG_TOMOYO_MAC_FOR_BINDPORT
-			if (ReadBindPolicy(head)) break;
-#endif
-			head->read_var2 = NULL; head->read_step = 5;
-		step5:
-#ifdef CONFIG_TOMOYO_MAC_FOR_CONNECTPORT
-			if (ReadConnectPolicy(head)) break;
-#endif
-			head->read_var2 = NULL; head->read_step = 6;
-		step6:
-#ifdef CONFIG_TOMOYO_MAC_FOR_SIGNAL
-			if (ReadSignalPolicy(head)) break;
-#endif
-			head->read_var2 = NULL; head->read_step = 7;
-		step7:
 			if (io_printf(head, "\n")) break;
 		}
 		if (!domain) head->read_eof = 1;
@@ -616,7 +721,7 @@ static int ReadDomainPolicy(IO_BUFFER *head)
 
 #endif
 
-/*************************  SYSTEM POLICY HANDLER  *************************/
+/*************************  EXCEPTION POLICY HANDLER  *************************/
 
 #ifdef CONFIG_TOMOYO
 
@@ -625,34 +730,26 @@ static int ReadDomainPolicy(IO_BUFFER *head)
 
 static int AddExceptionPolicy(char *data, void **domain)
 {
+	int is_delete = 0;
 	if (!isRoot()) return -EPERM;
+	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
+		data += KEYWORD_DELETE_LEN;
+		is_delete = 1;
+	}
 	if (strncmp(data, KEYWORD_TRUST_DOMAIN, KEYWORD_TRUST_DOMAIN_LEN) == 0) {
-		return AddTrustedPatternPolicy(data + KEYWORD_TRUST_DOMAIN_LEN);
+		return AddTrustedPatternPolicy(data + KEYWORD_TRUST_DOMAIN_LEN, is_delete);
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
 	} else if (strncmp(data, KEYWORD_ALLOW_READ, KEYWORD_ALLOW_READ_LEN) == 0) {
-		return AddGloballyReadablePolicy(data + KEYWORD_ALLOW_READ_LEN);
+		return AddGloballyReadablePolicy(data + KEYWORD_ALLOW_READ_LEN, is_delete);
 #endif
 	} else if (strncmp(data, KEYWORD_INITIALIZER, KEYWORD_INITIALIZER_LEN) == 0) {
-		return AddInitializerPolicy(data + KEYWORD_INITIALIZER_LEN);
+		return AddInitializerPolicy(data + KEYWORD_INITIALIZER_LEN, is_delete);
+	} else if (strncmp(data, KEYWORD_AGGREGATOR, KEYWORD_AGGREGATOR_LEN) == 0) {
+		return AddAggregatorPolicy(data + KEYWORD_AGGREGATOR_LEN, is_delete);
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
 	} else if (strncmp(data, KEYWORD_FILE_PATTERN, KEYWORD_FILE_PATTERN_LEN) == 0) {
-		return AddPatternPolicy(data + KEYWORD_FILE_PATTERN_LEN);
+		return AddPatternPolicy(data + KEYWORD_FILE_PATTERN_LEN, is_delete);
 #endif
-	} else if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
-		data += KEYWORD_DELETE_LEN;
-		if (strncmp(data, KEYWORD_TRUST_DOMAIN, KEYWORD_TRUST_DOMAIN_LEN) == 0) {
-			return DelTrustedPatternPolicy(data + KEYWORD_TRUST_DOMAIN_LEN);
-#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
-		} else if (strncmp(data, KEYWORD_ALLOW_READ, KEYWORD_ALLOW_READ_LEN) == 0) {
-			return DelGloballyReadablePolicy(data + KEYWORD_ALLOW_READ_LEN);
-#endif
-		} else if (strncmp(data, KEYWORD_INITIALIZER, KEYWORD_INITIALIZER_LEN) == 0) {
-			return DelInitializerPolicy(data + KEYWORD_INITIALIZER_LEN);
-#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
-		} else if (strncmp(data, KEYWORD_FILE_PATTERN, KEYWORD_FILE_PATTERN_LEN) == 0) {
-			return DelPatternPolicy(data + KEYWORD_FILE_PATTERN_LEN);
-#endif
-		}
 	}
 	return -EINVAL;
 }
@@ -689,6 +786,9 @@ static int ReadExceptionPolicy(IO_BUFFER *head)
 			if (ReadInitializerPolicy(head)) break;
 			head->read_var2 = NULL; head->read_step = 4;
 		case 4:
+			if (ReadAggregatorPolicy(head)) break;
+			head->read_var2 = NULL; head->read_step = 5;
+		case 5:
 #ifdef CONFIG_TOMOYO_MAC_FOR_FILE
 			if (ReadPatternPolicy(head)) break;
 #endif
@@ -712,42 +812,28 @@ static int ReadExceptionPolicy(IO_BUFFER *head)
 
 static int AddSystemPolicy(char *data, void **domain)
 {
+	int is_delete = 0;
 	if (!isRoot()) return -EPERM;
+	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
+		data += KEYWORD_DELETE_LEN;
+		is_delete = 1;
+	}
 #ifdef CONFIG_SAKURA_RESTRICT_MOUNT
 	if (strncmp(data, KEYWORD_ALLOW_MOUNT, KEYWORD_ALLOW_MOUNT_LEN) == 0)
-		return AddMountPolicy(data + KEYWORD_ALLOW_MOUNT_LEN);
+		return AddMountPolicy(data + KEYWORD_ALLOW_MOUNT_LEN, is_delete);
 #endif
 #ifdef CONFIG_SAKURA_RESTRICT_UNMOUNT
 	if (strncmp(data, KEYWORD_DENY_UNMOUNT, KEYWORD_DENY_UNMOUNT_LEN) == 0)
-		return AddNoUmountPolicy(data + KEYWORD_DENY_UNMOUNT_LEN);
+		return AddNoUmountPolicy(data + KEYWORD_DENY_UNMOUNT_LEN, is_delete);
 #endif
 #ifdef CONFIG_SAKURA_RESTRICT_CHROOT
 	if (strncmp(data, KEYWORD_ALLOW_CHROOT, KEYWORD_ALLOW_CHROOT_LEN) == 0)
-		return AddChrootPolicy(data + KEYWORD_ALLOW_CHROOT_LEN);
+		return AddChrootPolicy(data + KEYWORD_ALLOW_CHROOT_LEN, is_delete);
 #endif
 #ifdef CONFIG_SAKURA_RESTRICT_AUTOBIND
 	if (strncmp(data, KEYWORD_DENY_AUTOBIND, KEYWORD_DENY_AUTOBIND_LEN) == 0)
-		return AddReservedPortPolicy(data + KEYWORD_DENY_AUTOBIND_LEN);
+		return AddReservedPortPolicy(data + KEYWORD_DENY_AUTOBIND_LEN, is_delete);
 #endif
-	if (strncmp(data, KEYWORD_DELETE, KEYWORD_DELETE_LEN) == 0) {
-		data += KEYWORD_DELETE_LEN;
-#ifdef CONFIG_SAKURA_RESTRICT_MOUNT
-		if (strncmp(data, KEYWORD_ALLOW_MOUNT, KEYWORD_ALLOW_MOUNT_LEN) == 0)
-			return DelMountPolicy(data + KEYWORD_ALLOW_MOUNT_LEN);
-#endif
-#ifdef CONFIG_SAKURA_RESTRICT_UNMOUNT
-		if (strncmp(data, KEYWORD_DENY_UNMOUNT, KEYWORD_DENY_UNMOUNT_LEN) == 0)
-			return DelNoUmountPolicy(data + KEYWORD_DENY_UNMOUNT_LEN);
-#endif
-#ifdef CONFIG_SAKURA_RESTRICT_CHROOT
-		if (strncmp(data, KEYWORD_ALLOW_CHROOT, KEYWORD_ALLOW_CHROOT_LEN) == 0)
-			return DelChrootPolicy(data + KEYWORD_ALLOW_CHROOT_LEN);
-#endif
-#ifdef CONFIG_SAKURA_RESTRICT_AUTOBIND
-		if (strncmp(data, KEYWORD_DENY_AUTOBIND, KEYWORD_DENY_AUTOBIND_LEN) == 0)
-			return DelReservedPortPolicy(data + KEYWORD_DENY_AUTOBIND_LEN);
-#endif
-	}
 	return -EINVAL;
 }
 
@@ -828,22 +914,18 @@ void CCS_LoadPolicy(const char *filename)
 	}
 	/* Now try to load policies. */
 	init_started = 1;
-	{
-		int i;
-		for (i = 0; ccs_control_array[i].keyword; i++) {
-			if (ccs_control_array[i].assertion_index == i) continue;
-			panic("%s: FATAL: Control array index is broken. Fix the kernel source.\n", __FUNCTION__);
-		}
-	}
 	LoadPolicyManager();
 #ifdef CONFIG_SAKURA
-	printk("SAKURA: 1.0.1 2005/11/25\n");
-	printk("SAKURA Linux is loading policy. Please wait.\n");
+	printk("SAKURA: 1.1.2   2006/06/02\n");
+	printk("SAKURA is loading policy. Please wait.\n");
 	LoadSystemPolicy();
 #endif
 #ifdef CONFIG_TOMOYO
-	printk("TOMOYO: 1.0.1 2005/11/25\n");
-	printk("TOMOYO Linux is loading policy. Please wait.\n");
+	printk("TOMOYO: 1.1.2   2006/06/02\n");
+	printk("TOMOYO is loading policy. Please wait.\n");
+#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+	LoadPermissionMapping();
+#endif
 	LoadExceptionPolicy();
 	if (skip_domain_policy_loading == 0) LoadDomainPolicy();
 	{
@@ -851,7 +933,7 @@ void CCS_LoadPolicy(const char *filename)
 		int domain_count = 0;
 		int acl_count = 0;
 		for (domain = &KERNEL_DOMAIN; domain; domain = domain->next) {
-			FILE_ACL_RECORD *ptr = domain->first_acl_ptr;
+			struct acl_info *ptr = domain->first_acl_ptr;
 			while (ptr) { ptr = ptr->next; acl_count++; };
 			domain_count++;
 		}
@@ -861,6 +943,147 @@ void CCS_LoadPolicy(const char *filename)
 	LoadProfile();
 	printk("Mandatory Access Control activated.\n");
 	sbin_init_started = 1;
+	ccs_log_level = KERN_WARNING;
+}
+
+
+/*************************  MAC Decision Delayer  *************************/
+
+static DECLARE_WAIT_QUEUE_HEAD(query_wait);
+
+static spinlock_t query_lock = SPIN_LOCK_UNLOCKED;
+
+typedef struct query_entry {
+	struct list_head list;
+	unsigned int serial;
+	char *query;
+	int answer;
+} QUERY_ENTRY;
+
+static LIST_HEAD(query_list);
+
+int CheckSupervisor(const char *fmt, ...)
+{
+	va_list args;
+	int error = -EPERM;
+	int i, len;
+	static unsigned int serial = 0;
+	QUERY_ENTRY *query_entry;
+	if (!CheckCCSFlags(CCS_MAX_ENFORCE_GRACE)) return -EPERM;
+	va_start(args, fmt);
+	len = vsnprintf((char *) &i, sizeof(i) - 1, fmt, args) + 16;
+	va_end(args);
+	if ((query_entry = kzalloc(sizeof(QUERY_ENTRY), GFP_KERNEL)) == NULL ||
+		(query_entry->query = kzalloc(len, GFP_KERNEL)) == NULL) goto out;
+	INIT_LIST_HEAD(&query_entry->list);
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	query_entry->serial = serial++;
+	i = snprintf(query_entry->query, len - 1, "Q%u\n", query_entry->serial);
+	va_start(args, fmt);
+	vsnprintf(query_entry->query + i, len - 1 - i, fmt, args);
+	va_end(args);
+	list_add_tail(&query_entry->list, &query_list);
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	for (i = 0; i < CheckCCSFlags(CCS_MAX_ENFORCE_GRACE); i++) {
+		wake_up(&query_wait);
+		set_current_state(TASK_INTERRUPTIBLE);
+		schedule_timeout(HZ);
+		if (query_entry->answer) break;
+	}
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	list_del(&query_entry->list);
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	switch (query_entry->answer) {
+	case 1:
+		/* Granted by administrator. */
+		error = 0;
+		break;
+	case 0:
+		/* Timed out. */
+		break;
+	default:
+		/* Rejected by administrator. */
+		break;
+	}
+ out: ;
+	/* query_entry->query is set to NULL by ReadQuery(). */
+	if (query_entry) kfree(query_entry->query);
+	kfree(query_entry);
+	return error;
+}
+
+static int PollQuery(struct file *file, poll_table *wait)
+{
+	int found = 0;
+	struct list_head *tmp;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	list_for_each(tmp, &query_list) {
+		QUERY_ENTRY *ptr = list_entry(tmp, QUERY_ENTRY, list);
+		if (!ptr->query) continue;
+		found = 1;
+		break;
+	}
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	if (found) return POLLIN | POLLRDNORM;
+	poll_wait(file, &query_wait, wait);
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	list_for_each(tmp, &query_list) {
+		QUERY_ENTRY *ptr = list_entry(tmp, QUERY_ENTRY, list);
+		if (!ptr->query) continue;
+		found = 1;
+		break;
+	}
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	if (found) return POLLIN | POLLRDNORM;
+	return 0;
+}
+
+static int ReadQuery(IO_BUFFER *head)
+{
+	struct list_head *tmp;
+	if (head->read_avail) return 0;
+	if (head->read_buf) {
+		kfree(head->read_buf); head->read_buf = NULL;
+		head->readbuf_size = 0;
+	}
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	list_for_each(tmp, &query_list) {
+		QUERY_ENTRY *ptr = list_entry(tmp, QUERY_ENTRY, list);
+		if (!ptr->query) continue;
+		head->readbuf_size = head->read_avail = strlen(ptr->query) + 1;
+		head->read_buf = ptr->query; ptr->query = NULL;
+		break;
+	}
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	return 0;
+}
+
+static int WriteAnswer(char *data, void **dummy) 
+{
+	struct list_head *tmp;
+	unsigned int serial, answer;
+	if (sscanf(data, "A%u=%u", &serial, &answer) != 2) return -EINVAL;
+	/***** CRITICAL SECTION START *****/
+	spin_lock(&query_lock);
+	list_for_each(tmp, &query_list) {
+		QUERY_ENTRY *ptr = list_entry(tmp, QUERY_ENTRY, list);
+		if (ptr->serial != serial) continue;
+		if (!ptr->answer) ptr->answer = answer;
+		break;
+	}
+	spin_unlock(&query_lock);
+	/***** CRITICAL SECTION END *****/
+	return 0;
 }
 
 /*************************  /proc INTERFACE HANDLER  *************************/
@@ -876,22 +1099,19 @@ static int ReadMemoryCounter(IO_BUFFER *head)
 
 int CCS_OpenControl(const int type, struct file *file)
 {
-	IO_BUFFER *head = (IO_BUFFER *) kmalloc(sizeof(IO_BUFFER), GFP_KERNEL);
+	IO_BUFFER *head = (IO_BUFFER *) kzalloc(sizeof(IO_BUFFER), GFP_KERNEL);
 	if (!head) return -ENOMEM;
-	memset(head, 0, sizeof(IO_BUFFER));
 	head->readbuf_size = PAGE_SIZE * 2;
-	if ((head->read_buf = kmalloc(head->readbuf_size, GFP_KERNEL)) == NULL) {
+	if ((head->read_buf = kzalloc(head->readbuf_size, GFP_KERNEL)) == NULL) {
 		kfree(head);
 		return -ENOMEM;
 	}
 	head->writebuf_size = PAGE_SIZE * 2;
-	if ((head->write_buf = kmalloc(head->writebuf_size, GFP_KERNEL)) == NULL) {
+	if ((head->write_buf = kzalloc(head->writebuf_size, GFP_KERNEL)) == NULL) {
 		kfree(head->read_buf);
 		kfree(head);
 		return -ENOMEM;
 	}
-	memset(head->read_buf, 0, head->readbuf_size);
-	memset(head->write_buf, 0, head->writebuf_size);
 	init_MUTEX(&head->read_sem);
 	init_MUTEX(&head->write_sem);
 	switch (type) {
@@ -904,12 +1124,6 @@ int CCS_OpenControl(const int type, struct file *file)
 		head->write = AddExceptionPolicy;
 		head->read = ReadExceptionPolicy;
 		break;
-	case CCS_POLICY_DELETEDOMAIN:
-		head->write = DeleteDomain;
-		break;
-	case CCS_POLICY_UPDATEDOMAIN:
-		head->write = UpdateDomain;
-		break;
 	case CCS_INFO_TRUSTEDPIDS:
 		head->read = ReadTrustedPIDs;
 		break;
@@ -924,6 +1138,14 @@ int CCS_OpenControl(const int type, struct file *file)
 		head->poll = PollRejectLog;
 		head->read = ReadRejectLog;
 		break;
+	case CCS_INFO_SELFDOMAIN:
+		head->read = ReadSelfDomain;
+		break;
+#ifdef CONFIG_TOMOYO_MAC_FOR_FILE
+	case CCS_INFO_MAPPING:
+		head->read = ReadPermissionMapping;
+		break;
+#endif
 #endif
 #ifdef CONFIG_SAKURA
 	case CCS_POLICY_SYSTEMPOLICY:
@@ -938,6 +1160,11 @@ int CCS_OpenControl(const int type, struct file *file)
 		head->write = SetStatus;
 		head->read = ReadStatus;
 		break;
+	case CCS_POLICY_QUERY:
+		head->poll = PollQuery;
+		head->write = WriteAnswer;
+		head->read = ReadQuery;
+		break;
 	}
 	file->private_data = head;
 	return 0;
@@ -979,19 +1206,18 @@ int CCS_ReadControl(struct file *file, char __user *buffer, const int buffer_len
 int CCS_WriteControl(struct file *file, const char __user *buffer, const int buffer_len)
 {
 	IO_BUFFER *head = (IO_BUFFER *) file->private_data;
-	void *ptr = NULL;
+	void *ptr;
 	int error = buffer_len;
 	int avail_len = buffer_len;
 	char *cp0 = head->write_buf;
 	if (!head->write) return -ENOSYS;
 	if (!access_ok(VERIFY_READ, buffer, buffer_len)) return -EFAULT;
+	if (!isRoot()) return -EPERM;
 	if (!IsPolicyManager()) {
-		printk("This program is not permitted to update policies.\n");
 		return -EPERM; /* Forbid updating policies for non manager programs. */
 	}
 	if (down_interruptible(&head->write_sem)) return -EINTR;
-	if (head->write_step) ptr = head->write_var1;
-	head->write_step = 1;
+	ptr = head->write_var1;
 	while (avail_len > 0) {
 		char c;
 		if (head->write_avail >= head->writebuf_size - 1) {
